@@ -5,7 +5,10 @@ import functools
 import inspect
 import logging
 import re
+import warnings
 
+from .compat import is_callable, is_string, u
+from . import utils
 
 class WorkflowError(Exception):
     """Base class for errors from the xworkflows module."""
@@ -157,13 +160,13 @@ def _setup_states(state_definitions, prev=()):
         if len(state_def) != 2:
             raise TypeError("The 'state' attribute of a workflow should be "
                 "a two-tuple of strings; got %r instead." % (state_def,))
-        else:
-            name, title = state_def
-            state = State(name, title)
-        if state in states:
+        name, title = state_def
+        state = State(name, title)
+        if any(st.name == name for st in states):
             # Replacing an existing state
-            states.remove(state)
-        states.append(state)
+            states = [state if st.name == name else st for st in states]
+        else:
+            states.append(state)
     return StateList(states)
 
 
@@ -182,7 +185,7 @@ def _setup_transitions(tdef, states, prev=()):
     for transition in tdef:
         if len(transition) == 3:
             (name, source, target) = transition
-            if isinstance(source, basestring) or isinstance(source, State):
+            if is_string(source) or isinstance(source, State):
                 source = [source]
             source = [states[src] for src in source]
             target = states[target]
@@ -191,17 +194,105 @@ def _setup_transitions(tdef, states, prev=()):
             raise TypeError("Elements of the 'transition' attribute of a "
                 "workflow should be three-tuples; got %r instead." % (transition,))
 
-        at = None
-        for i, prev_tr in enumerate(trs):
-            if tr.name == prev_tr.name:
-                at = i
-        if at is not None:
-            trs[at] = tr
+        if any(prev_tr.name == tr.name for prev_tr in trs):
+            # Replacing an existing state
+            trs = [tr if prev_tr.name == tr.name else prev_tr for prev_tr in trs]
         else:
             trs.append(tr)
     return TransitionList(trs)
 
 
+HOOK_BEFORE = 'before'
+HOOK_AFTER = 'after'
+HOOK_CHECK = 'check'
+HOOK_ON_ENTER = 'on_enter'
+HOOK_ON_LEAVE = 'on_leave'
+
+
+class Hook(object):
+    """A hook to run when a transition occurs.
+
+    Attributes:
+        kind (str): the kind of hook
+        priority (int): the priority of the hook (higher values run first)
+        function (callable): the actual function to call
+        names (str list): name of the transitions or states to which the hook
+            relates. The special value '*' means 'applies to all transitions/
+            states'.
+
+    Hooks are sortable by descending priority and ascending function name.
+
+    Hook kinds are as follow:
+        - HOOK_BEFORE: run before the related transitions
+        - HOOK_AFTER: run after the related transitions
+        - HOOK_CHECK: run as part of pre-transition checks (return value matters)
+        - HOOK_ON_ENTER: run just after a transition entering a related state
+        - HOOK_ON_LEAVE: run just before a transition leaving from a related state
+    """
+
+    def __init__(self, kind, function, *names, **kwargs):
+        assert kind in (HOOK_BEFORE, HOOK_AFTER, HOOK_CHECK,
+            HOOK_ON_ENTER, HOOK_ON_LEAVE)
+
+        self.kind = kind
+        self.priority = kwargs.get('priority', 0)
+        self.function = function
+        self.names = names or ('*',)
+
+    def applies_to(self, transition, from_state=None):
+        """Whether this hook applies to the given transition/state.
+
+        Args:
+            transition (Transition): the transition to check
+            from_state (State or None): the state to check. If absent, the check
+                is 'might this hook apply to the related transition, given a
+                valid source state'.
+        """
+        if '*' in self.names:
+            return True
+        elif self.kind in (HOOK_BEFORE, HOOK_AFTER, HOOK_CHECK):
+            return transition.name in self.names
+        elif self.kind == HOOK_ON_ENTER:
+            return transition.target.name in self.names
+        elif from_state is None:
+            # Testing whether the hook may apply to at least one source of the
+            # transition
+            return any(src.name in self.names for src in transition.source)
+        else:
+            return from_state.name in self.names
+
+    def __call__(self, *args, **kwargs):
+        return self.function(*args, **kwargs)
+
+    def __eq__(self, other):
+        """Equality is based on priority, function and kind."""
+        if not isinstance(other, Hook):
+            return NotImplemented
+        return (self.priority == other.priority
+            and self.function == other.function
+            and self.kind == other.kind
+            and self.names == other.names
+        )
+
+    def __ne__(self, other):
+        if not isinstance(other, Hook):
+            return NotImplemented
+        return not (self == other)
+
+    def __cmp__(self, other):
+        """Compare hooks of the same kind."""
+        if not isinstance(other, Hook):
+            return NotImplemented
+        return cmp(
+            (other.priority, self.function.__name__),
+            (self.priority, other.function.__name__),
+        )
+
+    def __repr__(self):
+        return '<%s: %s %r>' % (
+            self.__class__.__name__, self.kind, self.function)
+
+
 class ImplementationWrapper(object):
     """Wraps a transition implementation.
 
@@ -214,30 +305,27 @@ class ImplementationWrapper(object):
         transition (Transition): the transition to perform
         workflow (Workflow): the workflow to which this is related.
 
-        check (callable): optional function to call along with state checks to
-            determine whether the transition is available.
-        before (callable): optional callable to call *before* performing the
-            transition.
+        hooks (Hook list): optional hooks to call during the transition
         implementation (callable): the code to invoke between 'before' and the
             state update.
-        after (callable): optional callable to call *after* changing the state
-            and logging the transition.
     """
 
     def __init__(self, instance, field_name, transition, workflow,
-            implementation, check=None, before=None, after=None):
+            implementation, hooks=None):
         self.instance = instance
         self.field_name = field_name
         self.transition = transition
         self.workflow = workflow
 
-        self.check = check
-        self.before = before
+        self.hooks = hooks or {}
         self.implementation = implementation
-        self.after = after
 
         self.__doc__ = implementation.__doc__
 
+    @property
+    def current_state(self):
+        return getattr(self.instance, self.field_name)
+
     def _pre_transition_checks(self):
         """Run the pre-transition checks."""
         current_state = getattr(self.instance, self.field_name)
@@ -246,15 +334,21 @@ class ImplementationWrapper(object):
                 "Transition '%s' isn't available from state '%s'." %
                 (self.transition.name, current_state.name))
 
-        if self.check is not None:
-            if not self.check(self.instance):
+        for check in self._filter_hooks(HOOK_CHECK):
+            if not check(self.instance):
                 raise ForbiddenTransition(
                     "Transition '%s' was forbidden by "
                     "custom pre-transition check." % self.transition.name)
 
+    def _filter_hooks(self, *hook_kinds):
+        """Filter a list of hooks, keeping only applicable ones."""
+        hooks = sum((self.hooks.get(kind, []) for kind in hook_kinds), [])
+        return sorted(hook for hook in hooks
+                      if hook.applies_to(self.transition, self.current_state))
+
     def _pre_transition(self, *args, **kwargs):
-        if self.before is not None:
-            self.before(self.instance, *args, **kwargs)
+        for hook in self._filter_hooks(HOOK_BEFORE, HOOK_ON_LEAVE):
+            hook(self.instance, *args, **kwargs)
 
     def _during_transition(self, *args, **kwargs):
         return self.implementation(self.instance, *args, **kwargs)
@@ -265,8 +359,8 @@ class ImplementationWrapper(object):
 
     def _post_transition(self, result, *args, **kwargs):
         """Performs post-transition actions."""
-        if self.after is not None:
-            self.after(self.instance, result, *args, **kwargs)
+        for hook in self._filter_hooks(HOOK_AFTER, HOOK_ON_ENTER):
+            hook(self.instance, result, *args, **kwargs)
 
     def __call__(self, *args, **kwargs):
         """Run the transition, with all checks."""
@@ -298,7 +392,7 @@ class ImplementationWrapper(object):
         return True
 
     def __repr__(self):
-        return "<%s for '%s' on '%s': %s>" % (self.__class__.__name__,
+        return "<%s for %r on %r: %r>" % (self.__class__.__name__,
             self.transition.name, self.field_name, self.implementation)
 
 
@@ -312,26 +406,22 @@ class ImplementationProperty(object):
         transition (Transition): the transition to perform
         workflow (Workflow): the workflow to which this is related.
 
-        check (callable): optional function to call along with state checks to
-            determine whether the transition is available.
-        before (callable): optional callable to call *before* performing the
-            transition.
+        hooks (Hook list): hooks to apply along with the transition.
         implementation (callable): the code to invoke between 'before' and the
             state update.
-        after (callable): optional callable to call *after* changing the state
-            and logging the transition.
     """
     def __init__(self, field_name, transition, workflow, implementation,
-            check=None, before=None, after=None):
+            hooks=None):
         self.field_name = field_name
         self.transition = transition
         self.workflow = workflow
-        self.check = check
-        self.before = before
+        self.hooks = hooks or {}
         self.implementation = implementation
-        self.after = after
         self.__doc__ = implementation.__doc__
 
+    def add_hook(self, hook):
+        self.hooks.setdefault(hook.kind, []).append(hook)
+
     def __get__(self, instance, owner):
         if instance is None:
             return self
@@ -343,8 +433,7 @@ class ImplementationProperty(object):
 
         return self.workflow.implementation_class(instance,
             self.field_name, self.transition, self.workflow,
-            self.implementation,
-            self.check, self.before, self.after)
+            self.implementation, self.hooks)
 
     def __repr__(self):
         return "<%s for '%s' on '%s': %s>" % (self.__class__.__name__,
@@ -379,13 +468,126 @@ class TransitionWrapper(object):
 
 def transition(trname='', field='', check=None, before=None, after=None):
     """Decorator to declare a function as a transition implementation."""
-    if callable(trname):
+    if is_callable(trname):
         raise ValueError("The @transition decorator should be called as "
             "@transition(['transition_name'], **kwargs)")
+    if check or before or after:
+        warnings.warn(
+            "The use of check=, before= and after= in @transition decorators is"
+            "deprecated in favor of @transition_check, @before_transition and "
+            "@after_transition decorators.",
+            DeprecationWarning,
+            stacklevel=2)
     return TransitionWrapper(trname, field=field,
         check=check, before=before, after=after)
 
 
+def _make_hook_dict(fun):
+    """Ensure the given function has a xworkflows_hook attribute.
+
+    That attribute has the following structure:
+    >>> {
+    ...     'before': [('state', <TransitionHook>), ...],
+    ... }
+    """
+    if not hasattr(fun, 'xworkflows_hook'):
+        fun.xworkflows_hook = {
+            HOOK_BEFORE: [],
+            HOOK_AFTER: [],
+            HOOK_CHECK: [],
+            HOOK_ON_ENTER: [],
+            HOOK_ON_LEAVE: [],
+        }
+    return fun.xworkflows_hook
+
+
+class _HookDeclaration(object):
+    """Base class for decorators declaring methods as transition hooks.
+
+    Args:
+        *names (str tuple): name of the states/transitions to bind to; use '*'
+            for 'all'
+        priority (int): priority of the hook, defaults to 0
+        field (str): name of the field to which the hooked transition relates
+
+    Usage:
+        >>> @_HookDeclaration('foo', 'bar', priority=4)
+        ... def my_hook(self):
+        ...   pass
+    """
+
+    def __init__(self, *names, **kwargs):
+        if not names:
+            names = ('*',)
+        self.names = names
+        self.priority = kwargs.get('priority', 0)
+        self.field = kwargs.get('field', '')
+
+    def _as_hook(self, func):
+        return Hook(self.hook_name, func, *self.names, priority=self.priority)
+
+    def __call__(self, func):
+        hook_dict = _make_hook_dict(func)
+        hooks = hook_dict[self.hook_name]
+        hooks.append((self.field, self._as_hook(func)))
+        return func
+
+
+class before_transition(_HookDeclaration):
+    """Decorates a method that should be called before a given transition.
+
+    Example:
+        >>> @before_transition('foobar')
+        ... def blah(self):
+        ...   pass
+    """
+    hook_name = HOOK_BEFORE
+
+
+class after_transition(_HookDeclaration):
+    """Decorates a method that should be called after a given transition.
+
+    Example:
+        >>> @after_transition('foobar')
+        ... def blah(self):
+        ...   pass
+    """
+    hook_name = HOOK_AFTER
+
+
+class transition_check(_HookDeclaration):
+    """Decorates a method that should be called after a given transition.
+
+    Example:
+        >>> @transition_check('foobar')
+        ... def blah(self):
+        ...   pass
+    """
+    hook_name = HOOK_CHECK
+
+
+class on_enter_state(_HookDeclaration):
+    """Decorates a method that should be used as a hook for a state.
+
+    Example:
+        >>> @on_enter_state('foo', 'bar')
+        ... def blah(self):
+        ...   pass
+    """
+    hook_name = HOOK_ON_ENTER
+
+
+class on_leave_state(_HookDeclaration):
+    """Decorates a method that should be used as a hook for a state.
+
+    Example:
+        >>> @on_leave_state('foo', 'bar')
+        ... def blah(self):
+        ...   pass
+    """
+    hook_name = HOOK_ON_LEAVE
+
+
 def noop(instance, *args, **kwargs):
     """NoOp function, ignores all arguments."""
     pass
@@ -400,7 +602,9 @@ class ImplementationList(object):
             name to the associated implementation.
         workflow (Workflow): the related workflow
         transitions_at (dict(str => str)): maps a transition name to the
-            name of the attribute holding the related transition.
+            name of the attribute holding the related implementation.
+        custom_implems (str set): list of transition names for which a custom
+            implementation has been defined.
     """
 
     def __init__(self, state_field, workflow):
@@ -408,6 +612,19 @@ class ImplementationList(object):
         self.workflow = workflow
         self.implementations = {}
         self.transitions_at = {}
+        self.custom_implems = set()
+
+    def load_parent_implems(self, parent_implems):
+        """Import previously defined implementations.
+
+        Args:
+            parent_implems (ImplementationList): List of implementations defined
+                in a parent class.
+        """
+        for trname, attr, implem in parent_implems.get_custom_implementations():
+            self.implementations[trname] = implem
+            self.transitions_at[trname] = attr
+            self.custom_implems.add(trname)
 
     def add_implem(self, transition, attribute, function, **kwargs):
         """Add an implementation.
@@ -428,6 +645,7 @@ class ImplementationList(object):
             **kwargs)
         self.implementations[transition.name] = implem
         self.transitions_at[transition.name] = attribute
+        return implem
 
     def should_collect(self, value):
         """Decide whether a given value should be collected."""
@@ -446,7 +664,9 @@ class ImplementationList(object):
             if self.should_collect(value):
                 transition = self.workflow.transitions[value.trname]
 
-                if value.trname in self.implementations:
+                if (value.trname in self.implementations
+                    and value.trname in self.custom_implems
+                    and name != self.transitions_at[value.trname]):
                     # We already have an implementation registered.
                     other_implem_at = self.transitions_at[value.trname]
                     raise ValueError(
@@ -454,14 +674,49 @@ class ImplementationList(object):
                         "%s for transition %s, which is already implemented "
                         "at %s." % (name, value, transition, other_implem_at))
 
-                self.add_implem(transition, name, value.func,
-                    check=value.check, before=value.before, after=value.after)
+                implem = self.add_implem(transition, name, value.func)
+                self.custom_implems.add(transition.name)
+                if value.check:
+                    implem.add_hook(Hook(HOOK_CHECK, value.check))
+                if value.before:
+                    implem.add_hook(Hook(HOOK_BEFORE, value.before))
+                if value.after:
+                    implem.add_hook(Hook(HOOK_AFTER, value.after))
+
+    def get_custom_implementations(self):
+        """Retrieve a list of cutom implementations.
+
+        Yields:
+            (str, str, ImplementationProperty) tuples: The name of the attribute
+                an implementation lives at, the name of the related transition,
+                and the related implementation.
+        """
+        for trname in self.custom_implems:
+            attr = self.transitions_at[trname]
+            implem = self.implementations[trname]
+            yield (trname, attr, implem)
 
     def add_missing_implementations(self):
         for transition in self.workflow.transitions:
             if transition.name not in self.implementations:
                 self.add_implem(transition, transition.name, noop)
 
+    def register_hooks(self, cls):
+        for field, value in utils.iterclass(cls):
+            if is_callable(value) and hasattr(value, 'xworkflows_hook'):
+                self.register_function_hooks(value)
+
+    def register_function_hooks(self, func):
+        """Looks at an object method and registers it for relevent transitions."""
+        for hook_kind, hooks in func.xworkflows_hook.items():
+            for field_name, hook in hooks:
+                if field_name and field_name != self.state_field:
+                    continue
+                for transition in self.workflow.transitions:
+                    if hook.applies_to(transition):
+                        implem = self.implementations[transition.name]
+                        implem.add_hook(hook)
+
     def _may_override(self, implem, other):
         """Checks whether an ImplementationProperty may override an attribute."""
         if isinstance(other, ImplementationProperty):
@@ -495,16 +750,12 @@ class ImplementationList(object):
             attrs[attrname] = implem
         return attrs
 
-    def transform(self, attrs, add_missing=True):
+    def transform(self, attrs):
         """Perform all actions on a given attribute dict."""
         self.collect(attrs)
-        if add_missing:
-            self.add_missing_implementations()
+        self.add_missing_implementations()
         self.fill_attrs(attrs)
 
-    def __repr__(self):
-        return '<%s: %r>' % (self.__class__.__name__, self._implems.values())
-
 
 class WorkflowMeta(type):
     """Base metaclass for all Workflows.
@@ -563,8 +814,8 @@ class Workflow(object):
         try:
             instance_repr = unicode(repr(instance), 'utf8', 'ignore')
         except (UnicodeEncodeError, UnicodeDecodeError):
-            instance_repr = u'<bad repr>'
-        logger.info(u'%s performed transition %s.%s (%s -> %s)', instance_repr,
+            instance_repr = u("<bad repr>")
+        logger.info(u("%s performed transition %s.%s (%s -> %s)"), instance_repr,
             self.__class__.__name__, transition.name, from_state.name,
             transition.target.name)
 
@@ -662,6 +913,8 @@ class WorkflowEnabledMeta(type):
     - one class attribute for each the attached workflows,
     - a '_workflows' attribute, a dict mapping each field_name to the related
         Workflow,
+    - a '_xworkflows_implems' attribute, a dict mapping each field_name to a
+        dict of related ImplementationProperty.
     - one class attribute for each transition for each attached workflow
     """
 
@@ -670,9 +923,6 @@ class WorkflowEnabledMeta(type):
         """Attach a workflow to the attribute list (create a StateProperty)."""
         attrs[field_name] = StateProperty(state_field.workflow, field_name)
 
-        implems = ImplementationList(field_name, state_field.workflow)
-        implems.transform(attrs)
-
     @classmethod
     def _find_workflows(mcs, attrs):
         """Finds all occurrences of a workflow in the attributes definitions.
@@ -687,13 +937,63 @@ class WorkflowEnabledMeta(type):
                 workflows[attribute] = StateField(value)
         return workflows
 
-    def __new__(mcs, name, bases, attrs):
-        workflows = mcs._find_workflows(attrs)
-        for field_name, state_field in workflows.items():
-            mcs._add_workflow(field_name, state_field, attrs)
+    @classmethod
+    def _add_transitions(mcs, field_name, workflow, attrs, implems=None):
+        """Collect and enhance transition definitions to a workflow.
 
+        Modifies the 'attrs' dict in-place.
+
+        Args:
+            field_name (str): name of the field transitions should update
+            workflow (Workflow): workflow we're working on
+            attrs (dict): dictionary of attributes to be updated.
+            implems (ImplementationList): Implementation list from parent
+                classes (optional)
+
+        Returns:
+            ImplementationList: The new implementation list for this field.
+        """
+        new_implems = ImplementationList(field_name, workflow)
+        if implems:
+            new_implems.load_parent_implems(implems)
+        new_implems.transform(attrs)
+
+        return new_implems
+
+    @classmethod
+    def _register_hooks(mcs, cls, implems):
+        for implem_list in implems.values():
+            implem_list.register_hooks(cls)
+
+    def __new__(mcs, name, bases, attrs):
+        # Map field_name => StateField
+        workflows = {}
+        # Map field_name => ImplementationList
+        implems = {}
+
+        # Collect workflows and implementations from parents
+        for base in reversed(bases):
+            if hasattr(base, '_workflows'):
+                workflows.update(base._workflows)
+                implems.update(base._xworkflows_implems)
+
+        workflows.update(mcs._find_workflows(attrs))
+
+        # Update attributes with workflow descriptions, and collect
+        # implementation declarations.
+        for field, state_field in workflows.items():
+            mcs._add_workflow(field, state_field, attrs)
+
+            implems[field] = mcs._add_transitions(
+                field, state_field.workflow, attrs, implems.get(field))
+
+        # Set specific attributes for children
         attrs['_workflows'] = workflows
-        return super(WorkflowEnabledMeta, mcs).__new__(mcs, name, bases, attrs)
+        attrs['_xworkflows_implems'] = implems
+
+        cls = super(WorkflowEnabledMeta, mcs).__new__(mcs, name, bases, attrs)
+        mcs._register_hooks(cls, implems)
+        return cls
 
 
 class BaseWorkflowEnabled(object):
